from typing import Optional, Tuple
from PIL import Image
from rich.console import Console, ConsoleOptions, RenderResult
from rich.measure import Measurement
from rich.segment import Segment
from rich.style import Style
from rich.color import Color


class ImageSection:
    """Renders the specified section of an image.

    Args:
        image (Image.Image): PIL image to render.
        x_interval (tuple[int, int], optional). Range of x-values to render [low, high).
            Defaults to [0, width).
        y_interval (tuple[int, int], optional). Range of y-values to render [low, high).
            Can start at -1. Defaults to [0, height).
    """

    def __init__(
        self,
        image: Image.Image,
        x_interval: Optional[tuple[int, int]] = None,
        y_interval: Optional[tuple[int, int]] = None,
    ):
        self.image = image

        w, h = image.size
        self.x_interval = x_interval
        if x_interval is None:
            self.x_interval = (0, w)

        self.y_interval = y_interval
        if y_interval is None:
            self.y_interval = (0, h)

        self.segments: dict[tuple[int, int], Segment] = {}

    def __rich_measure__(
        self, console: Console, options: ConsoleOptions
    ) -> Measurement:
        width = self.x_interval[1] - self.x_interval[0]
        return Measurement(width, width)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        _, h = self.image.size
        newline = Segment("\n", Style.null())
        segments = []
        for y in range(*self.y_interval, 2):
            for x in range(*self.x_interval):
                position = (x, y)

                # Check if we've already computed the segment for this position
                if position not in self.segments:
                    upper = None
                    if y >= 0:
                        upper = Color.from_rgb(*self.image.getpixel(position))

                    lower = None
                    if y < h:
                        lower = Color.from_rgb(*self.image.getpixel((x, y + 1)))

                    # Render each pixel as a half-character
                    if upper is None:
                        segment = Segment("▄", Style(color=lower))
                    else:
                        segment = Segment("▀", Style(color=upper, bgcolor=lower))

                    # Cache segment for next render
                    self.segments[position] = segment

                # Use cached segment
                segments.append(self.segments[position])

            segments.append(newline)

        print("rendered image")
        return segments
